# Emmy version 1.0.0
# EmmyInterpreter.py
# by Antoine Rauzy
# Copyrights (c) 2019 AltaRica Association
#
# This file is part of Emmy, a pedagogical implementation of graphs as a modeling tool.
# Emmy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Emmy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Emmy.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creator
# 3) Commands
# 4) Command compute
#   4.1) Decision diagrams
#   4.2) Minimal cutsets
#   4.3) Compute probability
# 5) Command print
#  5.1) Literal indices
#  5.2) Nodes
#  5.3) Products
#  5.4) Orders
#  5.5) Size
# 6) Miscellaneous

# 1) Imported modules
# -------------------

from S2MLCore import *

from EmmyDecisionDiagramBuilder import *
from EmmyStochasticSimulator import *

# 2) Creator
# ----------

class EmmyInterpreter(S2MLInterpreter):
  def __init__(self, engine):
    S2MLInterpreter.__init__(self, engine)
    self._decisionDiagramBuilder = self._engine.GetDecisionDiagramBuilder()
    self._stochasticSimulator = self._engine.GetStochasticSimulator()

  def InitializeEnvironmentVariables(self):
    self.SetValueOfEnvironmentVariable("seed", S2MLConstant.NewInteger(12345))
    self.SetValueOfEnvironmentVariable("number-of-tries", S2MLConstant.NewInteger(0))
    self.SetValueOfEnvironmentVariable("output", S2MLConstant.NewString("results.txt"))
    self.SetValueOfEnvironmentVariable("mode", S2MLConstant.NewSymbol("write"))

# 3) Commands
# -----------

  def _ParseCommand(self):
    commandName = self.LookUpArgument()
    if commandName=="compute":
      commandName = self.GetArgument()
      self._errors += self._ParseCommandCompute()
    elif commandName=="print":
      commandName = self.GetArgument()
      self._errors += self._ParseCommandPrint()
    else:
      S2MLInterpreter._ParseCommand(self)

# 4) Command compute
# ------------------

  def _ParseCommandCompute(self):
    what = self.GetArgument()
    if what=="decision-diagrams":
      return self._ParseCommandComputeDecisionDiagramsOfObservers()
    if what=="minimal-cutsets":
      return self._ParseCommandComputeMinimalCutsetsOfObserver()
    if what=="probabilities":
      return self._ParseCommandComputeProbabilitiesOfObservers()
    self._RaiseError("line %d\nSyntax error\n" % self._lineNumber)
    return 1

# 4.1) Decision diagrams
# ----------------------

  def _ParseCommandComputeDecisionDiagramsOfObservers(self):
    targetModel = self._GetTargetModel()
    if targetModel==None:
      return 1
    targetBlock = self._ParseTargetBlockOption(targetModel)
    if targetBlock==None:
      return 1
    targetDDName = self._ParseDecisionDiagramNameOption()
    if targetDDName==None:
      return 1
    self._decisionDiagramBuilder = self._engine.GetDecisionDiagramBuilder()
    error = self._decisionDiagramBuilder.ComputeDecisionDiagramsOfObservers(targetBlock, targetDDName)
    return error

# 4.2) Minimal cutsets
# --------------------

  def _ParseCommandComputeMinimalCutsetsOfObserver(self):
    targetModel = self._GetTargetModel()
    if targetModel==None:
      return 1
    targetBlock = self._ParseTargetBlockOption(targetModel)
    if targetBlock==None:
      return 1
    targetElement = self._ParseTargetElementOption(targetBlock)
    if targetElement==None:
      return 1
    sourceDDHandle = self._ParseDecisionDiagramHandleOption(targetElement)
    if sourceDDHandle==None:
      return 1
    targetDDName = self._ParseDecisionDiagramNameOption()
    if targetDDName==None:
      return 1
    self._decisionDiagramBuilder = self._engine.GetDecisionDiagramBuilder()
    error = self._decisionDiagramBuilder.ComputeMinimalCutsetsOfObserver(targetBlock, targetElement, sourceDDHandle, targetDDName)
    return error

# 4.3) Probability
# ----------------

  def _ParseCommandComputeProbabilitiesOfObservers(self):
    targetModel = self._GetTargetModel()
    if targetModel==None:
      return 1
    targetBlock = self._ParseTargetBlockOption(targetModel)
    if targetBlock==None:
      return 1
    targetDDName = self._ParseDecisionDiagramNameOption()
    if targetDDName==None:
      return 1
    missionTimes = self.GetFloatList()
    if missionTimes==None:
      return 1

    self._stochasticSimulator = self._engine.GetStochasticSimulator()
    self._stochasticSimulator.SetSeed(self.GetValueOfEnvironmentVariable("seed").GetValue())
    self._stochasticSimulator.SetNumberOfTries(self.GetValueOfEnvironmentVariable("number-of-tries").GetValue())
    self._stochasticSimulator.SetOutputFileName(self.GetValueOfEnvironmentVariable("output").GetValue())
    self._stochasticSimulator.SetOutputFileOpeningMode(self.GetValueOfEnvironmentVariable("mode").GetValue())
    profile = self._engine.GetDefaultIndicatorProfile()
    targetBlock.SetIndicatorProfileOfObservers(profile)

    while True:
      argument = self.LookUpArgument()
      if argument==None:
        break
      if argument=="seed":
        intValue = self.GetIntOption("seed", 12345)
        self._stochasticSimulator.SetSeed(intValue)
      elif argument=="output":
        stringValue = self.GetStringOption("output", "results.txt")
        self._stochasticSimulator.SetOutputFileName(stringValue)
      elif argument=="mode":
        stringValue = self.GetModeOption("write")
        self._stochasticSimulator.SetOutputFileOpeningMode(stringValue)
      else:
        observerName = self.LookUpArgument()
        observer = targetBlock.LookForIdentifiedElement(observerName, JanosModel.IsObserver)
        if observer==None:
          self._RaiseError("line %d\nNo observer %s\n" % (self._lineNumber, observerName))
          return 1
        argument = self.LookUpArgument()
        if argument==None or argument!=":":
          self._RaiseError("line %d\nA colon ':' was expected\n" % self._lineNumber)
          return 1
        profileName = self.LookUpArgument()
        if profileName==None:
          self._RaiseError("line %d\nA profile name was expected\n" % self._lineNumber)
          return 1
        profile = self._engine.GetIndicatorProfile(profileName)
        if profile==None:
          self._RaiseError("line %d\nNo indicator profile %s\n" % (self._lineNumber, profileName))
          return 1
        observer.SetIndicatorProfile(profile)

    if self._stochasticSimulator.GetNumberOfTries()==0:
      self._stochasticSimulator.CalculateProbabilitiesOfObservers(targetBlock, targetDDName, missionTimes)
    else:
      self._stochasticSimulator.PerformStochasticSimulationOnProbabilitiesOfObservers(targetBlock, targetDDName, missionTimes)
    return 0

# 5) Command print
# ----------------

  def _ParseCommandPrint(self):
    what = self.LookUpArgument()
    if what=="literal-indices":
      what = self.GetArgument()
      return self._ParseCommandPrintLiteralIndices()
    if what=="nodes":
      what = self.GetArgument()
      return self._ParseCommandPrintDecisionDiagramNodes()
    if what=="products":
      what = self.GetArgument()
      return self._ParseCommandPrintDecisionDiagramProducts()
    if what=="orders":
      what = self.GetArgument()
      return self._ParseCommandPrintDecisionDiagramOrdersOfProducts()
    if what=="size":
      what = self.GetArgument()
      return self._ParseCommandPrintDecisionDiagramSize()
    # self.PutBackArgument()
    return S2MLInterpreter._ParseCommandPrint(self)


# 5.1) Literal indices
# --------------------

  def _ParseCommandPrintLiteralIndices(self):
    targetModel = self._GetTargetModel()
    if targetModel==None:
      return 1
    outputFileName = None
    outputMode = "w"
    while True:
      argument = self.LookUpArgument()
      if argument==None:
        break
      if argument=="output":
        outputFileName = self.GetStringOption("output", "results.txt")
      elif argument=="mode":
        outputMode = self.GetModeOption("write")
      else:
        self._RaiseError("Incorrect argument %s" % argument)
        return 1
    printer = self._engine.GetPrinter()
    if outputFileName==None:
      error = printer.PrintLiteralIndices(self._engine.GetOutputFile())
    else:
      outputFilePath = self._engine.GetFilePath(outputFileName)
      error = printer.ExportLiteralIndices(outputFilePath, outputMode)
    return error

# 5.2) Nodes
# ----------

  def _ParseCommandPrintDecisionDiagramNodes(self):
    targetModel = self._GetTargetModel()
    if targetModel==None:
      return 1
    targetBlock = self._ParseTargetBlockOption(targetModel)
    if targetBlock==None:
      return 1
    targetElement = self._ParseTargetElementOption(targetBlock)
    if targetElement==None:
      return 1
    targetDDHandle = self._ParseDecisionDiagramHandleOption(targetElement)
    if targetDDHandle==None:
      return 1
    outputFileName = None
    outputMode = "w"
    while True:
      argument = self.LookUpArgument()
      if argument==None:
        break
      if argument=="output":
        outputFileName = self.GetStringOption("output", "results.txt")
      elif argument=="mode":
        outputMode = self.GetModeOption("write")
      else:
        self._RaiseError("Incorrect argument %s" % argument)
        return 1
    printer = self._engine.GetPrinter()
    if outputFileName==None:
      error = printer.PrintDecisionDiagramNodes(targetElement, targetDDHandle, self._engine.GetOutputFile())
    else:
      outputFilePath = self._engine.GetFilePath(outputFileName)
      error = printer.ExportDecisionDiagramNodes(targetElement, targetDDHandle, outputFilePath, outputMode)
    return error

# 5.3) Products
# -------------

  def _ParseCommandPrintDecisionDiagramProducts(self):
    targetModel = self._GetTargetModel()
    if targetModel==None:
      return 1
    targetBlock = self._ParseTargetBlockOption(targetModel)
    if targetBlock==None:
      return 1
    targetElement = self._ParseTargetElementOption(targetBlock)
    if targetElement==None:
      return 1
    targetDDHandle = self._ParseDecisionDiagramHandleOption(targetElement)
    if targetDDHandle==None:
      return 1
    printOrders = False
    printProbabilities = False
    missionTime = 0.0
    outputFileName = None
    outputMode = "w"
    while True:
      argument = self.LookUpArgument()
      if argument==None:
        break
      if argument=="print-orders":
        printOrders = self.GetBoolOption("print-orders", False)
      elif argument=="print-probabilities":
        printProbabilities = self.GetBoolOption("print-probabilities", False)
      elif argument=="mission-time":
        missionTime = self.GetFloatOption("mission-time", 0.0)
      elif argument=="output":
        outputFileName = self.GetStringOption("output", "results.txt")
      elif argument=="mode":
        outputMode = self.GetModeOption("write")
      else:
        self._RaiseError("Incorrect argument %s" % argument)
        return 1
    if printProbabilities:
      self._stochasticSimulator = self._engine.GetStochasticSimulator()
      self._stochasticSimulator.CalculateProbabilitiesOfLiterals(targetBlock, targetDDHandle.GetName(), missionTime)
    printer = self._engine.GetPrinter()
    if outputFileName==None:
      error = printer.PrintDecisionDiagramProducts(targetElement, targetDDHandle, printOrders, printProbabilities, missionTime, self._engine.GetOutputFile())
    else:
      outputFilePath = self._engine.GetFilePath(outputFileName)
      error = printer.ExportDecisionDiagramProducts(targetElement, targetDDHandle, printOrders, printProbabilities, missionTime, outputFilePath, outputMode)
    return error

# 5.4) Orders
# -----------

  def _ParseCommandPrintDecisionDiagramOrdersOfProducts(self):
    targetModel = self._GetTargetModel()
    if targetModel==None:
      return 1
    targetBlock = self._ParseTargetBlockOption(targetModel)
    if targetBlock==None:
      return 1
    targetElement = self._ParseTargetElementOption(targetBlock)
    if targetElement==None:
      return 1
    targetDDHandle = self._ParseDecisionDiagramHandleOption(targetElement)
    if targetDDHandle==None:
      return 1
    outputFileName = None
    outputMode = "w"
    while True:
      argument = self.LookUpArgument()
      if argument==None:
        break
      if argument=="output":
        outputFileName = self.GetStringOption("output", "results.txt")
      elif argument=="mode":
        outputMode = self.GetModeOption("write")
      else:
        self._RaiseError("Incorrect argument %s" % argument)
        return 1
    printer = self._engine.GetPrinter()
    if outputFileName==None:
      error = printer.PrintDecisionDiagramOrdersOfProducts(targetElement, targetDDHandle, self._engine.GetOutputFile())
    else:
      outputFilePath = self._engine.GetFilePath(outputFileName)
      error = printer.ExportDecisionDiagramOrdersOfProducts(targetElement, targetDDHandle, outputFilePath, outputMode)
    return error

# 5.5) Size
# ---------

  def _ParseCommandPrintDecisionDiagramSize(self):
    targetModel = self._GetTargetModel()
    if targetModel==None:
      return 1
    targetBlock = self._ParseTargetBlockOption(targetModel)
    if targetBlock==None:
      return 1
    targetElement = self._ParseTargetElementOption(targetBlock)
    if targetElement==None:
      return 1
    targetDDHandle = self._ParseDecisionDiagramHandleOption(targetElement)
    if targetDDHandle==None:
      return 1
    outputFileName = None
    outputMode = "w"
    while True:
      argument = self.LookUpArgument()
      if argument==None:
        break
      if argument=="output":
        outputFileName = self.GetStringOption("output", "results.txt")
      elif argument=="mode":
        outputMode = self.GetModeOption("write")
      else:
        self._RaiseError("Incorrect argument %s" % argument)
        return 1
    printer = self._engine.GetPrinter()
    if outputFileName==None:
      error = printer.PrintDecisionDiagramSize(targetElement, targetDDHandle, self._engine.GetOutputFile())
    else:
      outputFilePath = self._engine.GetFilePath(outputFileName)
      error = printer.ExportDecisionDiagramSize(targetElement, targetDDHandle, outputFilePath, outputMode)
    return error

# 6) Miscellaneous
# ----------------

  def _GetTargetModel(self):
    targetModel = self._engine.GetTargetModel()
    if targetModel==None:
      self._RaiseError("No instantiated model")
      return None
    return targetModel

  def _ParseTargetBlockOption(self, targetModel):
    if targetModel==None:
      self._RaiseError("line %d\nNo target model\n" % self._lineNumber)
      return None
    blockName = self.GetArgument()
    if blockName==None:
      self._RaiseError("line %d\nA block name was expected\n" % self._lineNumber)
      return None
    targetBlock = targetModel.LookForBlock(blockName)
    if targetBlock==None:
      self._RaiseError("No block %s in instantiated model" % blockName)
      return None
    return targetBlock

  def _ParseTargetElementOption(self, targetBlock):
    if targetBlock==None:
      self._RaiseError("line %d\nNo target block\n" % self._lineNumber)
      return
    elementName = self.GetArgument()
    if elementName==None:
      self._RaiseError("line %d\nA variable/observer name was expected\n" % self._lineNumber)
      return None
    targetElement = targetBlock.LookForIdentifiedElement(elementName)
    if targetElement==None or (targetElement.GetType()!=EmmyType.VARIABLE and targetElement.GetType()!=EmmyType.OBSERVER):
      self._RaiseError("No variable/observer %s in instantiated model" % elementName)
      return None
    return targetElement

  def _ParseDecisionDiagramNameOption(self):
    decisionDiagramName = self.GetArgument()
    if decisionDiagramName==None:
      self._RaiseError("line %d\nA decision diagram name was expected\n" % self._lineNumber)
      return None
    return decisionDiagramName

  def _ParseDecisionDiagramHandleOption(self, element):
    decisionDiagramName = self._ParseDecisionDiagramNameOption()
    if decisionDiagramName==None:
      return None
    if element.GetType()==EmmyType.VARIABLE:
      fdsVariable = element.GetFDSVariable()
      if fdsVariable==None:
        return None
      ddHandle = fdsVariable.LookForDDHandle(decisionDiagramName)
    elif element.GetType()==EmmyType.OBSERVER:
      fdsObserver = element.GetFDSObserver()
      if fdsObserver==None:
        return None
      ddHandle = fdsObserver.LookForDDHandle(decisionDiagramName)
    else:
      ddHandle = None
    if ddHandle==None:
      self._RaiseError("No decision diagram %s associated with variable/observer %s" % (decisionDiagramName, element.GetIdentifier()))
      return None
    return ddHandle
