# S2MLCore version 1.0.0
# S2MLEngine.py
# by Antoine Rauzy
# Copyrights (c) 2018 AltaRica Association
#
# This file is part of S2MLCore.
# S2MLCore is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# S2MLCore is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with S2MLCore.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creator
# 3) Settings
# 4) Methods to import, instantiate and export models
# 5) Management of environment variables
# 6) Management of indicator profiles

# 1) Imported modules
# -------------------

import pathlib

from S2MLModels import *
from S2MLInputsOutputs import *
from S2MLStatistics import *

from S2MLRewriter import *
from S2MLInterpreter import *

# 2) Creator
# ----------

class S2MLEngine:
  DEFAULT_INDICATOR_PROFILE_NAME = "profile1"

  # Execution statuses
  COMPLETED = 0
  ACTIVE = 1
  CANCELLED = 2

  def __init__(self):
    self._currentWorkingDirectory = pathlib.Path('.')
    self._lexer = None
    self._parser = None
    self._printer = None
    self._expressionManager = None
    self._flattener = None
    self._checker = None
    self._rewriter = None
    self._interpreter = None
    self._sourceModel = None
    self._targetModel = None
    self._targetBlock = None
    self._outputFile = sys.stdout
    self._numberOfErrors = 0
    self._errorFile = sys.stderr
    self._indicatorProfileManager = S2MLIndicatorProfileManager()
    profile = self._indicatorProfileManager.NewProfile(S2MLEngine.DEFAULT_INDICATOR_PROFILE_NAME)
    profile.SetMean(True)
    profile.SetStandardDeviation(True)
    profile.SetConfidenceInterval90(False)
    profile.SetConfidenceInterval95(True)
    profile.SetConfidenceInterval99(False)
    profile.SetExtrema(True)
    profile.SetQuantiles(False)
    profile.SetDistribution(False)
    profile.SetCumulativeDistribution(False)
    profile.SetBins(False)
    profile.SetBootstrapping(False)
    self._firstFreeMatrixIndex = 0
    self._matrix = None
    self._executionStatus = S2MLEngine.COMPLETED

# 3) Settings
# -----------

  def CreateTools(self):
    self._lexer = S2MLLexer(self)
    self._parser = S2MLParser(self)
    self._expressionManager = S2MLStochasticExpressionManager(self)
    self._printer = S2MLPrinter(self)
    self._flattener = S2MLFlattener(self)
    self._checker = S2MLChecker(self)
    self._rewriter = S2MLRewriter(self)
    self._interpreter = S2MLInterpreter(self)

  def GetCurrentWorkingDirectory(self):
    return self._currentWorkingDirectory
  def SetCurrentWorkingDirectory(self, directoryName):
    self._currentWorkingDirectory = self.GetFilePath(directoryName)
  def GetFilePath(self, fileName):
    filePath = pathlib.Path(fileName)
    if not filePath.is_absolute():
      filePath = self._currentWorkingDirectory / filePath
    filePath = filePath.resolve()
    return filePath

  def GetLexer(self):
    return self._lexer
  def GetParser(self):
    return self._parser
  def GetExpressionManager(self):
    return self._expressionManager
  def GetPrinter(self):
    return self._printer
  def GetFlattener(self):
    return self._flattener
  def GetChecker(self):
    return self._checker
  def GetInterpreter(self):
    return self._interpreter
  def GetMatrix(self):
    return self._matrix

  def OpenSession(self):
    self.CreateTools()
    return 0

  def CloseSession(self):
    self._outputFile.flush()
    self._errorFile.flush()
    return 0

  def GetOutputFile(self):
    return self._outputFile
  def SetOutputFile(self, outputFile):
    self._outputFile = outputFile

  def GetErrorFile(self):
    return self._errorFile
  def SetErrorFile(self, errorFile):
    self._errorFile = errorFile

  def GetNumberOfErrors(self):
    return self._numberOfErrors

  def RaiseError(self, message):
    self._numberOfErrors += 1
    self._errorFile.write("%s\n" % message)
    self._errorFile.flush()

  def GetExecutionStatus(self):
    return self._executionStatus
  def SetExecutionStatus(self, status):
    self._executionStatus = status

# 4) Methods to import, instantiate and export models
# ---------------------------------------------------

  # To be overloaded
  def NewModel(self):
    return S2MLModel()

  def ResetModel(self):
    self._sourceModel = self.NewModel()
    self._targetModel = None
    self._targetBlock = None
    self._numberOfErrors = 0
    return 0

  def GetSourceModel(self):
    return self._sourceModel
  def GetTargetModel(self):
    return self._targetModel

  def ImportModel(self, fileName):
    if self._sourceModel==None:
      self.ResetModel()
    if self._sourceModel==None:
      return 1
    self._targetModel = None
    self._targetBlock = None
    filePath = self.GetFilePath(fileName)
    numberOfErrors = self._parser.ImportFile(self._sourceModel, filePath)
    if numberOfErrors!=0:
      self.ResetModel()
    return numberOfErrors

  def LoadScript(self, fileName):
    self._numberOfErrors = 0
    filePath = self.GetFilePath(fileName)
    self._interpreter.LoadScript(str(filePath))
    return self._numberOfErrors

  def ParseCommand(self, commandText):
    self._numberOfErrors = 0
    self._interpreter.ParseCommand(commandText)
    return self._numberOfErrors

  def ExportModel(self, fileName, mode="w"):
    if self._sourceModel==None:
      self.RaiseError("No model to export")
      return 1
    filePath = self.GetFilePath(fileName)
    error = self._printer.ExportModel(self._sourceModel, filePath, mode)
    return error

  def PrintModel(self):
    if self._sourceModel==None:
      self.RaiseError("No model to export")
      return 1
    self._printer.PrintModel(self._sourceModel, self._outputFile)
    return 0

  def InstantiateModel(self):
    if self._sourceModel==None:
      self.RaiseError("No model to check")
      return 1
    self._targetModel = self._flattener.InstantiateAndFlattenModel(self._sourceModel)
    self._targetBlock = None
    if self._targetModel==None:
      return 1
    error = self._checker.CheckModel(self._targetModel)
    if error!=0:
      self._targetModel = None
      return error
    error = self._rewriter.PostProcessModel(self._targetModel)
    if error!=0:
      self._targetModel = None
      return error
    self._targetBlock = self._targetModel.GetBlock(0)
    return 0

  def ExportInstantiatedModel(self, fileName, mode="w"):
    if self._targetModel==None:
      self.RaiseError("No model to export")
      return 1
    filePath = self.GetFilePath(fileName)
    self._printer.ExportModel(self._targetModel, filePath, mode)
    return 0

  def PrintInstantiatedModel(self):
    if self._targetModel==None:
      self.RaiseError("No model to export")
      return 1
    self._printer.PrintModel(self._targetModel, self._outputFile)
    return 0

  def SetTargetBlock(self, blockName):
    if self._targetModel==None:
      self.RaiseError("No instantiated model")
      return 1
    self._targetBlock = self._targetModel.LookForBlock(blockName)
    if self._targetBlock==None:
      self.RaiseError("No block %s in instantiated model" % blockName)
      return 1
    return 0

# 5) Management of environment variables
# --------------------------------------

  def ExportEnvironmentVariables(self, fileName, mode="w"):
    filePath = self.GetFilePath(fileName)
    error = self._printer.ExportEnvironmentVariables(self._interpreter.GetEnvironmentVariables(), filePath, mode)
    return error

  def PrintEnvironmentVariables(self):
    self._printer.PrintEnvironmentVariables(self._interpreter.GetEnvironmentVariables(), self._outputFile)
    return 0

  def ExportEnvironmentVariable(self, identifier, fileName, mode="w"):
    value = self._interpreter.GetValueOfEnvironmentVariable(identifier)
    if value==None:
      self.RaiseError("No variable %s" % identifier)
      return 1
    filePath = self.GetFilePath(fileName)
    error = self._printer.ExportEnvironmentVariable(identifier, value, filePath, mode)
    return error

  def PrintEnvironmentVariable(self, identifier):
    value = self._interpreter.GetValueOfEnvironmentVariable(identifier)
    if value==None:
      self.RaiseError("No variable %s" % identifier)
      return 1
    self._printer.PrintEnvironmentVariable(identifier, value, self._outputFile)
    return 0

# 6) Management of indicator profiles
# -----------------------------------

  def GetIndicatorProfile(self, sourceName):
    return self._indicatorProfileManager.GetProfile(sourceName)

  def GetDefaultIndicatorProfile(self):
    profile = self._indicatorProfileManager.GetProfile(S2MLEngine.DEFAULT_INDICATOR_PROFILE_NAME)
    if profile!=None:
      return profile
    profile = self._indicatorProfileManager.GetFirstProfile()
    if profile!=None:
      return profile
    return self._indicatorProfileManager.GetProfile(S2MLEngine.DEFAULT_INDICATOR_PROFILE_NAME)

  def NewIndicatorProfile(self, sourceName):
    sourceProfile = self._indicatorProfileManager.GetProfile(sourceName)
    if sourceProfile!=None:
      self.RaiseError("Indicator profile %s already exist" % sourceName)
      return 1
    self._indicatorProfileManager.NewProfile(sourceName)
    return 0

  def CloneIndicatorProfile(self, sourceName, targetName):
    sourceProfile = self._indicatorProfileManager.GetProfile(sourceName)
    if sourceProfile==None:
      self.RaiseError("No indicator profile %s" % sourceName)
      return 1
    targetProfile = self._indicatorProfileManager.GetProfile(targetName)
    if targetProfile!=None:
      self.RaiseError("Indicator profile %s already exist" % targetName)
      return 1
    self._indicatorProfileManager.CloneProfile(sourceProfile, targetName)
    return 0

  def DeleteIndicatorProfile(self, sourceName):
    sourceProfile = self._indicatorProfileManager.GetProfile(sourceName)
    if sourceProfile==None:
      self.RaiseError("No indicator profile %s" % sourceName)
      return 1
    self._indicatorProfileManager.DeleteProfile(sourceProfile)
    return 0

  def SetIndicatorProfileOptionMean(self, sourceName, value):
    sourceProfile = self._indicatorProfileManager.GetProfile(sourceName)
    if sourceProfile==None:
      self.RaiseError("No indicator profile %s" % sourceName)
      return 1
    sourceProfile.SetMean(value)
    return 0
  def SetIndicatorProfileOptionStandardDeviation(self, sourceName, value):
    sourceProfile = self._indicatorProfileManager.GetProfile(sourceName)
    if sourceProfile==None:
      self.RaiseError("No indicator profile %s" % sourceName)
      return 1
    sourceProfile.SetStandardDeviation(value)
    return 0
  def SetIndicatorProfileOptionConfidenceInterval90(self, sourceName, value):
    sourceProfile = self._indicatorProfileManager.GetProfile(sourceName)
    if sourceProfile==None:
      self.RaiseError("No indicator profile %s" % sourceName)
      return 1
    sourceProfile.SetConfidenceInterval90(value)
    return 0
  def SetIndicatorProfileOptionConfidenceInterval95(self, sourceName, value):
    sourceProfile = self._indicatorProfileManager.GetProfile(sourceName)
    if sourceProfile==None:
      self.RaiseError("No indicator profile %s" % sourceName)
      return 1
    sourceProfile.SetConfidenceInterval95(value)
    return 0
  def SetIndicatorProfileOptionConfidenceInterval99(self, sourceName, value):
    sourceProfile = self._indicatorProfileManager.GetProfile(sourceName)
    if sourceProfile==None:
      self.RaiseError("No indicator profile %s" % sourceName)
      return 1
    sourceProfile.SetConfidenceInterval99(value)
    return 0
  def SetIndicatorProfileOptionExtrema(self, sourceName, value):
    sourceProfile = self._indicatorProfileManager.GetProfile(sourceName)
    if sourceProfile==None:
      self.RaiseError("No indicator profile %s" % sourceName)
      return 1
    sourceProfile.SetExtrema(value)
    return 0
  def SetIndicatorProfileOptionQuantiles(self, sourceName, value):
    sourceProfile = self._indicatorProfileManager.GetProfile(sourceName)
    if sourceProfile==None:
      self.RaiseError("No indicator profile %s" % sourceName)
      return 1
    sourceProfile.SetQuantiles(value)
    return 0
  def SetIndicatorProfileOptionNumberOfQuantiles(self, sourceName, value):
    sourceProfile = self._indicatorProfileManager.GetProfile(sourceName)
    if sourceProfile==None:
      self.RaiseError("No indicator profile %s" % sourceName)
      return 1
    sourceProfile.SetNumberOfQuantiles(value)
    return 0
  def SetIndicatorProfileOptionDistribution(self, sourceName, value):
    sourceProfile = self._indicatorProfileManager.GetProfile(sourceName)
    if sourceProfile==None:
      self.RaiseError("No indicator profile %s" % sourceName)
      return 1
    sourceProfile.SetDistribution(value)
    return 0
  def SetIndicatorProfileOptionCumulativeDistribution(self, sourceName, value):
    sourceProfile = self._indicatorProfileManager.GetProfile(sourceName)
    if sourceProfile==None:
      self.RaiseError("No indicator profile %s" % sourceName)
      return 1
    sourceProfile.SetCumulativeDistribution(value)
    return 0
  def SetIndicatorProfileOptionNumberOfPoints(self, sourceName, value):
    sourceProfile = self._indicatorProfileManager.GetProfile(sourceName)
    if sourceProfile==None:
      self.RaiseError("No indicator profile %s" % sourceName)
      return 1
    sourceProfile.SetNumberOfPoints(value)
    return 0
  def SetIndicatorProfileOptionBins(self, sourceName, value):
    sourceProfile = self._indicatorProfileManager.GetProfile(sourceName)
    if sourceProfile==None:
      self.RaiseError("No indicator profile %s" % sourceName)
      return 1
    sourceProfile.SetBins(value)
    return 0
  def SetIndicatorProfileOptionNumberOfBins(self, sourceName, value):
    sourceProfile = self._indicatorProfileManager.GetProfile(sourceName)
    if sourceProfile==None:
      self.RaiseError("No indicator profile %s" % sourceName)
      return 1
    sourceProfile.SetNumberOfBins(value)
    return 0
  def SetIndicatorProfileOptionNumberOfStoredValues(self, sourceName, value):
    sourceProfile = self._indicatorProfileManager.GetProfile(sourceName)
    if sourceProfile==None:
      self.RaiseError("No indicator profile %s" % sourceName)
      return 1
    sourceProfile.SetNumberOfStoredValues(value)
    return 0
  def SetIndicatorProfileOptionBootstrapping(self, sourceName, value):
    sourceProfile = self._indicatorProfileManager.GetProfile(sourceName)
    if sourceProfile==None:
      self.RaiseError("No indicator profile %s" % sourceName)
      return 1
    sourceProfile.SetBootstrapping(value)
    return 0
  def SetIndicatorProfileOptionNumberOfSamples(self, sourceName, value):
    sourceProfile = self._indicatorProfileManager.GetProfile(sourceName)
    if sourceProfile==None:
      self.RaiseError("No indicator profile %s" % sourceName)
      return 1
    sourceProfile.SetNumberOfSamples(value)
    return 0
  def SetIndicatorProfileOptionSamplingRatio(self, sourceName, value):
    sourceProfile = self._indicatorProfileManager.GetProfile(sourceName)
    if sourceProfile==None:
      self.RaiseError("No indicator profile %s" % sourceName)
      return 1
    sourceProfile.SetSamplingRatio(value)
    return 0

  def ExportIndicatorProfiles(self, fileName, mode="w"):
    error = self._printer.ExportIndicatorProfiles(self._indicatorProfileManager.GetProfiles(), fileName, mode)
    return error

  def PrintIndicatorProfiles(self):
    self._printer.PrintIndicatorProfiles(self._indicatorProfileManager.GetProfiles(), self._outputFile)
    return 0

  def ExportIndicatorProfile(self, identifier, fileName, mode="w"):
    profile = self._indicatorProfileManager.GetProfile(identifier)
    if profile==None:
      self.RaiseError("No profile %s" % identifier)
      return 1
    filePath = self.GetFilePath(fileName)
    error = self._printer.ExportIndicatorProfile(profile, filePath, mode)
    return error

  def PrintIndicatorProfile(self, identifier):
    profile = self._indicatorProfileManager.GetProfile(identifier)
    if profile==None:
      self.RaiseError("No profile %s" % identifier)
      return 1
    self._printer.PrintIndicatorProfile(profile, self._outputFile)
    return 0
